/*
 * exempi - capi.cpp
 *
 * Copyright (C) 2007 Hubert Figuiere
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, 
 * USA
 */
/** @brief the C API glue code */

#include <exempi/xmp.h>

#include "xmppacket.h"
#include "xmptree.h"
#include "xmpvalue.h"
#include "xmpxmlparser.h"
#include "iterwrap.h"

using namespace Exempi;
using Capi::IterWrap;

extern "C" {


XmpPtr xmp_new(const char *buffer, size_t len)
{
	Internal::XmpXmlParser parser(buffer, len);
	XmpPacket *xmp = new XmpPacket(&parser);
	if (!xmp->isValid()) {
		delete xmp;
		xmp = NULL;
	}
	return reinterpret_cast<XmpPtr>(xmp);
}

const char * xmp_get_property(XmpPtr xmp, const char *schema, 
															const char *name)
{
	XmpPacket *xmppacket = reinterpret_cast<XmpPacket*>(xmp);

	const XmpTree & tree = xmppacket->tree();
	XmpTree::const_iterator iter = tree.dataStore().find(schema);
	if (iter != tree.end()) {
		const XmpSchema & currentSchema = iter->second;
		XmpSchema::const_iterator iter2 = 
			currentSchema.properties.find(name);
		if (iter2 != currentSchema.end()) {
			try {
				return boost::any_cast<std::string>(iter2->second->value).c_str();
			}
			catch(...) {
			}
		}
	}
	return NULL;
}

void xmp_free(XmpPtr xmp)
{
	XmpPacket *xmppacket = reinterpret_cast<XmpPacket*>(xmp);
	delete xmppacket;
}

XmpTreePtr xmp_get_tree(XmpPtr xmp)
{
	XmpPacket *xmppacket = reinterpret_cast<XmpPacket*>(xmp);
	XmpTree & tree = xmppacket->tree();
	return reinterpret_cast<XmpTreePtr>(&tree);
}

XmpSchemaIterator xmp_tree_get_schema_iter(XmpTreePtr tree)
{
	XmpTree *xtree = reinterpret_cast<XmpTree*>(tree);
	IterWrap<XmpTree> *iter = new IterWrap<XmpTree>(*xtree);
	return reinterpret_cast<XmpSchemaIterator>(iter);
}


int xmp_schema_iter_next(XmpSchemaIterator iter)
{
	IterWrap<XmpTree> *i = reinterpret_cast<IterWrap<XmpTree> *>(iter);
	return i->next();
}


int xmp_schema_iter_is_valid(XmpSchemaIterator iter)
{
	IterWrap<XmpTree> *i = reinterpret_cast<IterWrap<XmpTree> *>(iter);
	return i->isValid();
}


const char* xmp_schema_iter_name(XmpSchemaIterator iter)
{
	IterWrap<XmpTree> *i = reinterpret_cast<IterWrap<XmpTree> *>(iter);	
	return i->data().first.c_str();
}


XmpSchemaPtr xmp_schema_iter_value(XmpSchemaIterator iter)
{
	IterWrap<XmpTree> *i = reinterpret_cast<IterWrap<XmpTree> *>(iter);	
	XmpSchema * s = &(i->data().second);
	return reinterpret_cast<XmpSchemaPtr>(s);
}


void xmp_schema_iter_free(XmpSchemaIterator iter)
{
	delete reinterpret_cast<IterWrap<XmpTree>*>(iter);
}



XmpPropertyIterator xmp_schema_get_property_iter(XmpSchemaPtr schema)
{
	XmpSchema *xschema = reinterpret_cast<XmpSchema*>(schema);
	IterWrap<XmpSchema> *iter = new IterWrap<XmpSchema>(*xschema);
	return reinterpret_cast<XmpPropertyIterator>(iter);
}

int xmp_property_iter_next(XmpPropertyIterator iter)
{
	IterWrap<XmpSchema> *i = reinterpret_cast<IterWrap<XmpSchema> *>(iter);
	return i->next();
}


int xmp_property_iter_is_valid(XmpPropertyIterator iter)
{
	IterWrap<XmpSchema> *i = reinterpret_cast<IterWrap<XmpSchema> *>(iter);
	return i->isValid();
}


const char* xmp_property_iter_name(XmpPropertyIterator iter)
{
	IterWrap<XmpSchema> *i = reinterpret_cast<IterWrap<XmpSchema> *>(iter);	
	return i->data().first.c_str();
}


XmpValuePtr xmp_property_iter_value(XmpPropertyIterator iter)
{
	IterWrap<XmpSchema> *i = reinterpret_cast<IterWrap<XmpSchema> *>(iter);	
	return reinterpret_cast<XmpValuePtr>(i->data().second.get());
}


void xmp_property_iter_free(XmpPropertyIterator iter)
{
	delete reinterpret_cast<IterWrap<XmpSchema>*>(iter);
}


const char* xmp_value_get_string(XmpValuePtr value)
{
	XmpValue *val = reinterpret_cast<XmpValue*>(value);
	try {
		return boost::any_cast<std::string>(val->value).c_str();
	}
	catch(const boost::bad_any_cast &) {
	}
	return NULL;
}

const char* xmp_value_get_field(XmpValuePtr value, const char *field)
{
	XmpValue *val = reinterpret_cast<XmpValue*>(value);
	
	return val->getField(field);
}

}
