/*
 * exempi - xmpxmlparser.cpp
 *
 * Copyright (C) 2007 Hubert Figuiere
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, 
 * USA
 */

#include <cstring>
#include <iostream>
#include <boost/any.hpp>

#include <libxml/encoding.h>

#include <exempi/xmpconsts.h>

#include "xmpxmlparser.h"


namespace Exempi {
	namespace Internal {

		XmpXmlParser::XmpXmlParser(const char * buffer, size_t len)
			: m_reader(NULL),
				m_depth(-1),
				m_context(),
				m_currentSchema(),
				m_currentProperty()
		{
			m_reader = xmlReaderForMemory(buffer, len,
																		"/", 
																		xmlGetCharEncodingName(XML_CHAR_ENCODING_UTF8),
																		0);
		}
		
		
		XmpXmlParser::~XmpXmlParser()
		{
			xmlFreeTextReader(m_reader);
		}
		
		
		bool XmpXmlParser::processNode(XmpTree::DataStore & store)
		{
			int current_depth = xmlTextReaderDepth(m_reader);
			int type = xmlTextReaderNodeType(m_reader);
			const xmlChar* n = xmlTextReaderConstLocalName(m_reader);
			const xmlChar* ns = xmlTextReaderConstNamespaceUri(m_reader);
			switch(current_depth) 
			{
			case 0: // x:xmpmeta
				if ((type == XML_READER_TYPE_ELEMENT) ||
						(type == XML_READER_TYPE_END_ELEMENT)) {
					if (strcmp((const char *)ns, NS_XMP_META) != 0) {
						return false;
					}
					if (strcmp((const char *)n, "xmpmeta") != 0) {
						return false;
					}
					if (!checkContext(type, CONTEXT_XMPMETA)) {
						return false;
					}
				}
				break;
			case 1: // rdf:RDF
				if ((type == XML_READER_TYPE_ELEMENT) ||
						(type == XML_READER_TYPE_END_ELEMENT)){
					if (strcmp((const char *)ns, NS_RDF) != 0) {
						return false;
					}
					if (strcmp((const char *)n, "RDF") != 0) {
						return false;
					}
					if (!checkContext(type, CONTEXT_RDF)) {
						return false;
					}
				}
				break;
			case 2: // rdf:Description
				if ((type == XML_READER_TYPE_ELEMENT) ||
						(type == XML_READER_TYPE_END_ELEMENT)) {
					if (strcmp((const char *)ns, NS_RDF) != 0) {
						return false;
					}
					if (strcmp((const char *)n, "Description") != 0) {
						return false;
					}
					if (!checkContext(type, CONTEXT_RDFDESCRIPTION)) {
						return false;
					}
					if (type == XML_READER_TYPE_END_ELEMENT) {
						m_currentSchema = store.end();
					}
					else {
						if (!xmlTextReaderHasAttributes(m_reader)) {
							return false;
						}
						const xmlChar *schema = NULL;
						const xmlChar *prefix = NULL;
						while(xmlTextReaderMoveToNextAttribute(m_reader) == 1) {
							const xmlChar *attrPrefix = xmlTextReaderConstPrefix(m_reader);
							if (strcmp((char*)attrPrefix, "xmlns") == 0) {
								schema = xmlTextReaderConstValue(m_reader);
								prefix = xmlTextReaderConstLocalName(m_reader);
								break;
							}
						}
						if (schema == NULL) {
							return false;
						}
						std::pair<XmpTree::DataStore::iterator, bool> i =
							store.insert(XmpTree::DataStore::value_type(
														 (const char*)schema,
														 XmpSchema((const char*)prefix))
								);
						// FIXME: what if it already exists?
						m_currentSchema = i.first;
//						std::cout << "currentSchema = " << schema << std::endl;
					}
				}
				break;
			case 3:
				if ((type == XML_READER_TYPE_ELEMENT) ||
						(type == XML_READER_TYPE_END_ELEMENT)) {
					if (!checkContext(type, CONTEXT_PROPERTY)) {
						return false;
					}
					
					if (type == XML_READER_TYPE_ELEMENT) {
						const xmlChar *pfx = xmlTextReaderConstPrefix(m_reader);
						if (m_currentSchema->second.prefix == (const char*)pfx) {
							std::pair<XmpSchema::Properties::iterator, bool> i =
								m_currentSchema->second.properties.insert(
									XmpSchema::Properties::value_type((const char*)n,
																										XmpValue::Ptr(new XmpValue()))
									);
							m_currentProperty = i.first;
							// parse down the property
							bool success =  processProperty(m_currentProperty);
							// if we got a property then we reprocess the current
							// node
							if (success) {
								return processNode(store);
							}
						}
						else {
							std::cout << "unexpected prefix " << pfx << std::endl;
						}
					}
					else {
						// END_ELEMENT
						m_currentProperty = m_currentSchema->second.properties.end();
					}
				}
				break;
			default:
//				if (!checkContext(type, CONTEXT_RDFDESCRIPTION)) {
//						return false;
//				}
				break;
			}

			return true;
		}

		bool XmpXmlParser::processProperty(XmpSchema::Properties::iterator & prop)
		{
			if (m_context.top() != CONTEXT_PROPERTY) {
				std::cout << "wrong context to parse property" << std::endl;
				return false;
			}

			int ret = 0;
			bool isResource = false;
			// try to find the attribute that tells it is a  rdf:Resource
			ret = xmlTextReaderMoveToFirstAttribute(m_reader);
			while (ret == 1) {
				const xmlChar *n = xmlTextReaderConstLocalName(m_reader);
				const xmlChar *ns = xmlTextReaderConstNamespaceUri(m_reader);
				if (strcmp((const char *)ns, NS_RDF) == 0) {
					if (strcmp((const char *)n, "parseType") == 0) {
						const xmlChar *v = xmlTextReaderConstValue(m_reader);
						isResource = (strcmp((const char *)v, "Resource") == 0);
						prop->second->type = XmpValue::XMP_VALUE_TYPE_RESOURCE;
//						std::cout << "isResource " << isResource << std::endl;
						break;
					}
				}
				ret = xmlTextReaderMoveToNextAttribute(m_reader);
			}
			xmlTextReaderMoveToElement(m_reader);

			// parse the inner content
			int start_depth = xmlTextReaderDepth(m_reader);
			ret = xmlTextReaderRead(m_reader);
			if (ret == 1) {
				if(xmlTextReaderNodeType(m_reader) == XML_READER_TYPE_TEXT) {
					const xmlChar *value = xmlTextReaderConstValue(m_reader);
					if (value != NULL) {
						XmpValue::Ptr xmp_value(prop->second);
						xmp_value->type = XmpValue::XMP_VALUE_TYPE_SINGLE;
						xmp_value->value = std::string((const char*)value);
//						std::cout << "property " << prop->first << " is " 
//											<< value << std::endl;
					}
				}
				else {
					int depth = start_depth;
//					std::cout << "parsing value for property " 
//										<< prop->first << std::endl;
					do {
						depth = xmlTextReaderDepth(m_reader);
						if (depth > start_depth) {
							if (xmlTextReaderNodeType(m_reader) == XML_READER_TYPE_ELEMENT) {
								const xmlChar* n = xmlTextReaderConstLocalName(m_reader);
								const xmlChar* ns = xmlTextReaderConstNamespaceUri(m_reader);
								if (isResource) {
									if (m_currentSchema->first == (const char *)ns) {
//										std::cout << "found resource " << n << std::endl;
									}
								}
#if 0
								else if (strcmp((const char *)ns, NS_DIMENSIONS_TYPE) == 0) {
									std::cout << "ns = " << ns << std::endl;
									prop->second.type = XmpValue::XMP_VALUE_TYPE_DIMENSIONS;
									prop->second.value = XmpValue::Structure();
									XmpValue::Structure *stru = 
										boost::any_cast<XmpValue::Structure>(&prop->second.value);
									/*return*/ parseStructure(*stru);
								}
#endif
								else if (strcmp((const char *)ns, NS_RDF) != 0) {
									std::cout << "wrong namespace (1) " << ns << std::endl;
									return false;
								} 
								else {
									if(strcmp((const char *)n, "Bag") == 0) {
										prop->second->type = XmpValue::XMP_VALUE_TYPE_BAG;
										prop->second->value = XmpValue::Bag();
										XmpValue::Bag * bag = 
											boost::any_cast<XmpValue::Bag>(&(prop->second->value));
										parseBag(*bag);
									}
									else if(strcmp((const char *)n, "Seq") == 0) {
										prop->second->type = XmpValue::XMP_VALUE_TYPE_SEQ;
										prop->second->value = XmpValue::Seq();
										XmpValue::Seq * seq = 
											boost::any_cast<XmpValue::Seq>(&(prop->second->value));
										parseSeq(*seq);										
									}
									else if(strcmp((const char *)n, "Alt") == 0) {
										prop->second->type = XmpValue::XMP_VALUE_TYPE_ALT;
										prop->second->value = XmpValue::Alt();
										XmpValue::Alt * alt = 
											boost::any_cast<XmpValue::Alt>(&(prop->second->value));
										parseAlt(*alt);																			
									}
									else if(strcmp((const char *)n, "Description") == 0) {
										prop->second->type = XmpValue::XMP_VALUE_TYPE_STRUCTURE;
										prop->second->value = XmpValue::Structure();
										XmpValue::Structure *stru = 
											boost::any_cast<XmpValue::Structure>(&(prop->second->value));
										parseStructure(*stru);										
									}
//									std::cout << "found property elem " << n << std::endl;
								}
							}
							ret = xmlTextReaderRead(m_reader);
						}
					} while((depth > start_depth) && (ret == 1));
				}
			}
			return true;
		}


		template <typename T>
		bool XmpXmlParser::parseArray(T & array)
		{
//			std::cout << "parseArray()" << std::endl;
			int start_depth = xmlTextReaderDepth(m_reader);
			int depth = start_depth;
			int ret = xmlTextReaderRead(m_reader);
			do {
				depth = xmlTextReaderDepth(m_reader);
				if (xmlTextReaderNodeType(m_reader) == XML_READER_TYPE_ELEMENT) {
					const xmlChar* n = xmlTextReaderConstLocalName(m_reader);
					const xmlChar* ns = xmlTextReaderConstNamespaceUri(m_reader);
					if (strcmp((const char *)ns, NS_RDF) != 0) {
						std::cout << "wrong namespace (2) " << ns << std::endl;
						return false;
					} 
					if (strcmp((const char *)n, "li") == 0) {
						ret = xmlTextReaderRead(m_reader);
						if ((ret == 1) 
								&& (xmlTextReaderNodeType(m_reader) == XML_READER_TYPE_TEXT)) {
							array.insert(array.end(), (const char*)xmlTextReaderConstValue(m_reader));
						}
					}
				}
				ret = xmlTextReaderRead(m_reader);				
			} while((depth > start_depth) && (ret == 1));
			return true;
		}


		bool XmpXmlParser::parseBag(XmpValue::Bag & bag)
		{
			return parseArray<XmpValue::Bag>(bag);
		}


		bool XmpXmlParser::parseSeq(XmpValue::Seq & seq)
		{
			return parseArray<XmpValue::Seq>(seq);
		}


		bool XmpXmlParser::parseAlt(XmpValue::Alt & alt)
		{
			// find a we to refactor this code with findArray()
//			std::cout << "parseAlt()" << std::endl;
			int start_depth = xmlTextReaderDepth(m_reader);
			int depth = start_depth;
			int ret = xmlTextReaderRead(m_reader);
			do {
				depth = xmlTextReaderDepth(m_reader);
				if (xmlTextReaderNodeType(m_reader) == XML_READER_TYPE_ELEMENT) {
					const xmlChar* n = xmlTextReaderConstLocalName(m_reader);
					const xmlChar* ns = xmlTextReaderConstNamespaceUri(m_reader);
					if (strcmp((const char *)ns, NS_RDF) != 0) {
						std::cout << "wrong namespace (3) " << ns << std::endl;
						return false;
					} 
					if (strcmp((const char *)n, "li") == 0) {
						xmlChar *lang = xmlTextReaderGetAttribute(m_reader, (const xmlChar*)"xml:lang");
						if (lang != NULL) {
							ret = xmlTextReaderRead(m_reader);
							if ((ret == 1) 
									&& (xmlTextReaderNodeType(m_reader) 
											== XML_READER_TYPE_TEXT)) {
								alt[(const char*)lang] = (const char*)xmlTextReaderConstValue(m_reader);
							}
							xmlFree(lang);
						}
					}
				}
				ret = xmlTextReaderRead(m_reader);				
			} while((depth > start_depth) && (ret == 1));
			return true;
		}


		bool XmpXmlParser::parseStructure(XmpValue::Structure & stru)
		{
			int start_depth = xmlTextReaderDepth(m_reader);
			int depth = start_depth;
			int ret = xmlTextReaderRead(m_reader);
			do {
				depth = xmlTextReaderDepth(m_reader);
				if (xmlTextReaderNodeType(m_reader) == XML_READER_TYPE_ELEMENT) {
					const xmlChar* n = xmlTextReaderConstLocalName(m_reader);
					//const xmlChar* ns = xmlTextReaderConstNamespaceUri(m_reader);
					// TODO check the namespace match the struct definition
					ret = xmlTextReaderRead(m_reader);
					if ((ret == 1) 
							&& (xmlTextReaderNodeType(m_reader) == XML_READER_TYPE_TEXT)) {
						stru[(const char*)n] = (const char*)xmlTextReaderConstValue(m_reader);
					}
				}
				ret = xmlTextReaderRead(m_reader);				
			} while((depth > start_depth) && (ret == 1));
			return true;
		}


		bool XmpXmlParser::checkContext(int type, int ctx)
		{
			if (type == XML_READER_TYPE_ELEMENT) {
				m_context.push(ctx);
			}
			else if (type == XML_READER_TYPE_END_ELEMENT) {
				if (m_context.top() != ctx) {
					std::cout << "wrong context wanted = " << ctx
										<< " current =  " << m_context.top() 
										<< std::endl;
					return false;
				}
				else {
					m_context.pop();
				}
			}
			else {
				std::cout << "wrong element type" << std::endl;
				return false;
			}
			return true;
		}
		
		bool XmpXmlParser::parse(XmpTree::DataStore & store)
		{
			m_currentSchema = store.end();
			if (m_reader != NULL) {
				bool loaded = false;
				try {
					int ret = xmlTextReaderRead(m_reader);
					while (ret == 1) {
						if (!processNode(store)) {
							std::cout << "semantic parsing failed" << std::endl;
							break;
						}
						ret = xmlTextReaderRead(m_reader);
					}
					loaded = (ret == 0);
					if (!loaded) {
						std::cout << "error " << ret << std::endl;
					}
				}
				catch(...)
				{
					loaded = false;
				}
				return loaded;
			}
			return false;
		}

		
	}
}
